/***************************************************************************
                          regio.c  -  description
                             -------------------
    begin                : Tue Dec 28 1999
    copyright            : (C) 1999 by Eugene S. Weiss
    email                : eweiss@sas.upenn.edu
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

/* These functions read and write data to the scanner's conteol registers.
They are deliberately slowed somewhat (and should be more so if necessary
in order to match the scanner's slower controller chip */

#include "as6edriver.h"
#include "regdata.h"
#include "prototypes.h"


/*--------------------------------------------------------------------------*/
//Clears the dreaded epp timeout.  Adapted from the Linux kernel (parport_pc.c).

inline void clear_epp_timeout(int port)
{
    unsigned char status;

    if (!(inb(statusport(port)) & 0x01))
	return;			//no timeout

    else {
	inb(statusport(port));
	status = inb(statusport(port));
	outb(status | 0x01, statusport(port));	/* Some reset by writing 1 */
	outb(status & 0xfe, statusport(port));	/* Others by writing 0 */
	return;
    }
}



/*--------------------------------------------------------------------------*/
//sets the ECR (extended control register) of a modern ECP parallel port to
//the communication mode that we wish to use.  Also sets the MODE global variable
//so other routines can check the operating mode.  Should in most cases be
//followed by a line setting the scanner's primary register to the appropriate
//mode so that the i/o will work in that mode.

void set_mode(int port, unsigned char newecrmode, unsigned char newmode)
{
    unsigned char oldecrmode, frob, test, ctl;

    if (newecrmode == ECR_EPP) {
	clear_epp_timeout(port);
	ctl = inb_p(controlport(port));
	outb_p(((ctl & ~DIRECTION) | nINIT), controlport(port));
	write_register(port, primary_register, newmode);
    }

    oldecrmode = inb_p(port + 0x402);
    frob = oldecrmode & 0x1f;	//mask out first 3 bits
    outb_p((frob | newecrmode), (port + 0x402));
    test = inb_p(port + 0x402);
//printf("\nold = 0x%x  got back 0x%x" ,oldecrmode, test);  fflush(stdout);
    ECRMODE = newecrmode;
    MODE = newmode;
    if (newecrmode != ECR_EPP)
	write_register(port, primary_register, newmode);
}


/*--------------------------------------------------------------------------*/

//write's a byte (data) to the register at addr and reads it back to make
//sure timing problems haven't kept it from getting through

void write_byte(int port, unsigned char addr, unsigned char data)
{

    unsigned char ctl, rddata, counter = 0;

    rddata = data - 1;

    while (rddata != data) {
	if (MODE != EPP_MODE) {
	    //tell the scanner which epp register is to receive data. This
	    //routine simulates the epp protocol, so that is can be used in
	    //slower modes.
	    outb_p(addr, port);
	    ctl = inb_p(controlport(port));
	    ctl |= STROBE;
	    outb_p(ctl, controlport(port));
	    ctl |= SELECTIN;
	    outb_p(ctl, controlport(port));
	    ctl &= ~SELECTIN;
	    outb_p(ctl, controlport(port));
	    ctl &= ~STROBE;
	    outb_p(ctl, controlport(port));

	    //simulates an epp data write
	    outb_p(data, port);
	    ctl = inb_p(controlport(port));
	    ctl |= STROBE;
	    outb_p(ctl, controlport(port));
	    ctl |= AUTOFEED;
	    outb_p(ctl, controlport(port));
	    ctl &= ~AUTOFEED;
	    outb_p(ctl, controlport(port));
	    ctl &= ~STROBE;
	    outb_p(ctl, controlport(port));
	} else {
	    clear_epp_timeout(port);
	    outb_p(addr, eppaddressport(port));	//otherwise use a direct write
	    outb_p(data, eppdataport(port));
	    clear_epp_timeout(port);
	}

	//don't try to read the data back if the write is to take the scanner offline
	if (((addr == 0x01) && (data == 0x40)) || (counter > 250))
	    break;

	//read a byte back in the proper mode
	if (MODE == BYTE_MODE)
	    rddata = byte_me(port, addr);
	else if (MODE == NIBBLE_MODE)
	    rddata = nibble_in(port, addr);
	else if (MODE == EPP_MODE)
	    rddata = epp_in(port, addr);
	if ((rddata != data) && (addr != 0x1))
	    fprintf(LOGd,
		    "\n Register at 0x%x doesn't match, is 0x%x, should be 0x%x",
		    addr, rddata, data);
	counter++;
    }
}

/*--------------------------------------------------------------------------*/

//reads a character in nibble mode.  from Luigi Rizzo's ppscan.c.

unsigned char nibble_in(int port, unsigned char addr)
{

    unsigned char ctl, buf[2], c;
    int ix;

    //set register address
    outb_p(addr, port);
    ctl = inb_p(controlport(port));
    ctl |= STROBE;
    outb_p(ctl, controlport(port));
    ctl |= SELECTIN;
    outb_p(ctl, controlport(port));
    ctl &= ~SELECTIN;
    outb_p(ctl, controlport(port));
    ctl &= ~STROBE;
    outb_p(ctl, controlport(port));

#define nibble2char(s) (0xf & (((s & ~nACK) >> 3) | (~s & nBUSY) >> 4))
    //start nibbling
    ctl = inb_p(controlport(port));
    c = (ctl | nINIT) & ~(AUTOFEED | STROBE | SELECTIN);
    if (c != ctl)
	fprintf(LOGd, "\nnibble status bad: 0x%02x ought to be 0x%02x",
		ctl, c);

    for (ix = 0; ix < 2; ix++) {
	c |= AUTOFEED;		//host ready for nibble
	if (ix == 0)
	    c &= ~nINIT;	//tell scanner to start nibbling out
	outb_p(c, controlport(port));
	buf[ix] = inb_p(statusport(port));	//  printf("\nnibble = 0x%x",buf[ix]);
	c &= ~AUTOFEED;		//got the data
	c |= nINIT;		//keep going
	outb_p(c, controlport(port));

	buf[ix] = nibble2char(buf[ix]);
    }
    buf[0] = buf[0] | (buf[1] << 4);

    return buf[0];
}

/*--------------------------------------------------------------------------*/

//reads a character from a register in byte mode

unsigned char byte_me(int port, unsigned char addr)
{

    unsigned char data, ctl, oldctl;

    //set address
    outb_p(addr, port);
    ctl = inb_p(controlport(port));
    ctl |= STROBE;
    outb_p(ctl, controlport(port));
    ctl |= SELECTIN;
    outb_p(ctl, controlport(port));
    ctl &= ~SELECTIN;
    outb_p(ctl, controlport(port));
    ctl &= ~STROBE;
    outb_p(ctl, controlport(port));

    //read in a byte ieee 1284 bidir mode
    ctl = inb_p(controlport(port));
    oldctl = ctl;
    ctl |= (DIRECTION | AUTOFEED);
    outb_p(ctl, controlport(port));
    data = inb_p(port);
    outb_p(oldctl, controlport(port));

    return (data);
}

/*--------------------------------------------------------------------------*/
//Reads a register value in epp mode.

unsigned char epp_in(int port, unsigned char addr)
{

    unsigned char data;
    clear_epp_timeout(port);
    outb_p(addr, eppaddressport(port));
    data = inb_p(eppdataport(port));
    clear_epp_timeout(port);
    return data;
}

/*--------------------------------------------------------------------------*/

//writes a value to any register

void write_register(int port, int regnumber, unsigned short data)
{

    unsigned char datahigh, datalow, high, low, oldmode = 0, oldecrmode =
	0;

//some registers can't be written to properly except in nibble mode probably
//because writing a value of 0 to the primary register resets the scanner as
//well as setting the data output mode to nibble

    if ((20 < regnumber) && (regnumber < 24)) {
	oldmode = MODE;
	oldecrmode = ECRMODE;

	set_mode(port, ECR_NIBBLE, NIBBLE_MODE);
//        write_register(port,primary_register,NIBBLE_MODE);
    }

    low = REGISTER[regnumber].low;
    high = REGISTER[regnumber].high;

    datahigh = (data >> 8 & 0xff);	//change a 2 byte input into bytes
    datalow = (data & 0xff);

    write_byte(port, low, datalow);

    if (REGISTER[regnumber].type == WORD)
	write_byte(port, high, datahigh);

    //restore the port mode if we have changed it to nibble mode
    if ((20 < regnumber) && (regnumber < 24)) {
	set_mode(port, oldecrmode, oldmode);
    }
}

/*--------------------------------------------------------------------------*/

//reads the value of any register

unsigned short read_register(int port, int regnumber)
{
    unsigned char low, high;

    low = REGISTER[regnumber].low;
    high = REGISTER[regnumber].high;

    if (ECRMODE == ECR_NIBBLE)
	low = nibble_in(port, low);
    else if (ECRMODE == ECR_BYTE)
	low = byte_me(port, low);
    else if (ECRMODE == ECR_EPP)
	low = epp_in(port, low);

    if (REGISTER[regnumber].type == WORD) {
	if (ECRMODE == ECR_NIBBLE)
	    high = nibble_in(port, high);
	else if (ECRMODE == ECR_BYTE)
	    high = byte_me(port, high);
	else if (ECRMODE == ECR_EPP)
	    high = epp_in(port, high);
    }

    return ((high << 8) | low);
}
